/**
 *
 * \file          gyx1d.c
 * \version       1
 * \date          29-Dec-08
 * \Author        Xiaomeng
 *
 */


/*============================================================================*/
/*                   STANDARD INCLUDE FILES                                   */
/*============================================================================*/
#include <linux/i2c.h>

/*============================================================================*/
/*                   PROJECT INCLUDE FILES                                    */
/*============================================================================*/
#include "dvb_frontend.h"
#include "gyx1d.h"

/*============================================================================*/
/*                   MACRO DEFINITION                                         */
/*============================================================================*/
#define IF              36166666
/*============================================================================*/
/*                   TYPE DEFINITION                                          */
/*============================================================================*/
struct gyx1d_ctx {
    struct i2c_adapter*     i2c;
    u8                      i2c_addr;
};

/*============================================================================*/
/*                   PUBLIC VARIABLES DEFINITION                              */
/*============================================================================*/


/*============================================================================*/
/*                   STATIC VARIABLES DECLARATIONS                            */
/*============================================================================*/


/*============================================================================*/
/*                   STATIC FUNCTIONS DECLARATIONS                            */
/*============================================================================*/
static int 
gyx1d_read_status (
        struct gyx1d_ctx*  ctx
        );

/*============================================================================*/
/*                   PUBLIC FUNCTIONS DEFINITIONS                             */
/*============================================================================*/


/*============================================================================*/
/* gyx1d_release                                                          */
/*============================================================================*/
int 
gyx1d_release (
        struct dvb_frontend*    fe
        )
{
    struct gyx1d_ctx*  ctx = (struct gyx1d_ctx*) fe->tuner_priv;

    /* free memory */
    kfree(ctx);
    fe->tuner_priv = NULL;

    return 0;
}

/*============================================================================*/
/* gyx1d_init                                                            */
/*============================================================================*/
int 
gyx1d_init (
        struct dvb_frontend*    fe
        )
{

    struct gyx1d_ctx*  ctx = (struct gyx1d_ctx*) fe->tuner_priv;

    /* close gate */
    if (fe->ops.i2c_gate_ctrl)
        fe->ops.i2c_gate_ctrl(fe, 1);

    return 0;
}

/*============================================================================*/
/* gyx1d_pll_set                                                         */
/*============================================================================*/
int 
gyx1d_pll_set (
        struct dvb_frontend*            fe,
        struct dvb_frontend_parameters* params
        )
{
    u8 tune_data[]={0,0,0,0,0};
    unsigned long tunerFreq;
    unsigned long uFreqPll;
    int i = 0;
    int status_pll = 0;
    struct gyx1d_ctx*  ctx = (struct gyx1d_ctx*) fe->tuner_priv;
    struct i2c_msg tuner_write_msg = {.addr = ctx->i2c_addr, .flags = 0};

    //Add to frequencie the IF before set the tuner PLL
    tunerFreq = params->frequency + IF;

    //-------------
    // byte 0 and 1
    //-------------
    // DB1 & DB2
    uFreqPll = tunerFreq + 166666;
    uFreqPll /= 166666;

    tune_data[0] = (u8)(uFreqPll >> 8);
    tune_data[1] = (u8)uFreqPll;

    //--------
    // byte 2
    //--------
    // CB1
    tune_data[2] = 0x90;

    //--------
    // byte 3
    //--------
    // BB
    if (tunerFreq < (100000000+IF))
        tune_data[3] = 0x81;
    else if (tunerFreq < (145000000+IF))
        tune_data[3] = 0xC1;
    else if (tunerFreq < (300000000+IF))
        tune_data[3] = 0x82;
    else if (tunerFreq < (430000000+IF))
        tune_data[3] = 0xC2;
    else if (tunerFreq < (700000000+IF))
        tune_data[3] = 0x88;
    else 
        tune_data[3] = 0xC8;

    if(params->u.ofdm.bandwidth == BANDWIDTH_8_MHZ){
        tune_data[3] |= 0x10;
    }
    else{
        tune_data[3] &= 0xEF;
    }
    //--------
    // byte 4
    //--------
    // CB2
    tune_data[4] = 0xC3;

    //-------------------------------------
    // Now we send the params to the tuner
    //-------------------------------------
    tuner_write_msg.buf = tune_data;
    tuner_write_msg.len = 5;
    if (i2c_transfer(ctx->i2c, &tuner_write_msg, 1) != 1)
    {
        return -EIO;
    }

    msleep(1);

    i=0;
    do
    {
        status_pll = gyx1d_read_status(ctx);
    }while(++i < 100 && !(status_pll & 0x40));

    return 0;

}

/*============================================================================*/
/*                   STATIC FUNCTIONS DEFINTIONS                              */
/*============================================================================*/

/*============================================================================*/
/* gyx1d_read_status                                                     */
/*============================================================================*/
static int 
gyx1d_read_status (
        struct gyx1d_ctx*  ctx
        )
{
    u8 b1[]={0};
    struct i2c_msg msg[] = {{ .addr = ctx->i2c_addr, .flags = I2C_M_RD,
        .buf=b1, .len = 1}};

    if (i2c_transfer(ctx->i2c, msg, 1) != 1){
        return -EIO;
    }

    return b1[0];
}

/**
 * Standard tuner operations
 */
static struct dvb_tuner_ops gyx1d_tuner_ops = {
	.info =
	{
		.name = "LG Gyx1D",
		.frequency_min  = 174000000, /* TODO TBC */
		.frequency_max  = 864000000, /* TODO TBC */
		.frequency_step = 8000000
	},
  .set_params        = gyx1d_pll_set,
  .get_frequency     = 0,
  .get_bandwidth     = 0,
  .release           = gyx1d_release,
  .init              = gyx1d_init,
};

/*============================================================================*/
/* gyx1d_attach                                                          */
/*============================================================================*/
struct dvb_frontend * gyx1d_attach (
        struct dvb_frontend*    fe,
        u8 i2c_addr,
        struct i2c_adapter*     i2c
        )
{
    struct gyx1d_ctx*  ctx = NULL;

    /* allocate memory for the internal state */
    ctx = (struct gyx1d_ctx*)kmalloc(sizeof(struct gyx1d_ctx), GFP_KERNEL);
    if (ctx == NULL) {
        return NULL;
    }

    memset(ctx, 0, sizeof(struct gyx1d_ctx));

    ctx->i2c = i2c;
    ctx->i2c_addr = i2c_addr;
    
    memcpy(&fe->ops.tuner_ops, &gyx1d_tuner_ops, sizeof(gyx1d_tuner_ops));
    
    fe->tuner_priv = ctx;

    return fe;
}

EXPORT_SYMBOL(gyx1d_attach);
/*============================================================================*/
/*                     END OF FILE                                            */
/*============================================================================*/


