/**
 *
 * \file          td1316.c
 * \version       1
 * \date          10-Oct-06
 * \Author        Guillot B.
 *
 */


/*============================================================================*/
/*                   STANDARD INCLUDE FILES                                   */
/*============================================================================*/
#include <linux/i2c.h>

/*============================================================================*/
/*                   PROJECT INCLUDE FILES                                    */
/*============================================================================*/
#include "dvb_frontend.h"
#include "td1316.h"

/*============================================================================*/
/*                   MACRO DEFINITION                                         */
/*============================================================================*/
#define IF              36166000
/*============================================================================*/
/*                   TYPE DEFINITION                                          */
/*============================================================================*/
struct td1316_ctx {
    struct i2c_adapter*     i2c;
    u8                      i2c_addr;
    int                     is_td1316_model_mk2_3;
    enum td1316_type        type;
};

/*============================================================================*/
/*                   PUBLIC VARIABLES DEFINITION                              */
/*============================================================================*/


/*============================================================================*/
/*                   STATIC VARIABLES DECLARATIONS                            */
/*============================================================================*/


/*============================================================================*/
/*                   STATIC FUNCTIONS DECLARATIONS                            */
/*============================================================================*/
static int 
td1316_mk23_pll_set (
        struct dvb_frontend*            fe,
        struct dvb_frontend_parameters* params
        );

static int 
td1316_mk14_pll_set (
        struct dvb_frontend*            fe,
        struct dvb_frontend_parameters* params
        );

static int 
td1316_read_status (
        struct td1316_ctx*  ctx
        );

/*============================================================================*/
/*                   PUBLIC FUNCTIONS DEFINITIONS                             */
/*============================================================================*/


/*============================================================================*/
/* td1316_release                                                          */
/*============================================================================*/
int 
td1316_release (
        struct dvb_frontend*    fe
        )
{
    struct td1316_ctx*  ctx = (struct td1316_ctx*) fe->tuner_priv;

    /* free memory */
    kfree(ctx);
    fe->tuner_priv = NULL;

    return 0;
}

/*============================================================================*/
/* td1316_init                                                            */
/*============================================================================*/
int 
td1316_init (
        struct dvb_frontend*    fe
        )
{

    struct td1316_ctx*  ctx = (struct td1316_ctx*) fe->tuner_priv;
    int StatusByte;
    int DiffBits;

    /* if type = TDQG4, don't check if model is mk2_3 or mk1_4, force to mk1_4 */
    if (ctx->type == TDQG4)
    {
        ctx->is_td1316_model_mk2_3 = 0;
        return 0;
    }
    /* if type = DTT8D1T, don't check if model is mk2_3 or mk1_4, force to mk2_3 */
    if (ctx->type == DTT8D1T)
    {
        ctx->is_td1316_model_mk2_3 = 1;
        return 0;
    }
    /* if type = DCT4C2, don't check if model is mk2_3 or mk1_4, force to mk1_4 */
    if (ctx->type == DCT4C2)
    {
        ctx->is_td1316_model_mk2_3 = 0;
        return 0;
    }
    
    /* close gate */
    if (fe->ops.i2c_gate_ctrl)
		    fe->ops.i2c_gate_ctrl(fe, 1);
		/* read status */
    StatusByte  = td1316_read_status(ctx);
    /* re -open gate is done in dvb-core layer */
    if (StatusByte < 0)
    {
    	printk("%s : error reading status\n", __FUNCTION__);
    	return -1;
    }
    
    DiffBits    = (StatusByte & 0x30) >> 4;

    if ( DiffBits == 0x3 )
    {
        ctx->is_td1316_model_mk2_3 = 1;
    }
    else
    {
        ctx->is_td1316_model_mk2_3 = 0;

    }

    return 0;
}

/*============================================================================*/
/* td1316_pll_set                                                         */
/*============================================================================*/
int 
td1316_pll_set (
        struct dvb_frontend*            fe,
        struct dvb_frontend_parameters* params
        )
{
    int error_status;
    struct td1316_ctx*  ctx = (struct td1316_ctx*) fe->tuner_priv;

    if( ctx->is_td1316_model_mk2_3 )
    {
        error_status= td1316_mk23_pll_set (fe, params);
    }
    else
    {
        error_status= td1316_mk14_pll_set (fe, params);
    }
    return error_status;

}

/*============================================================================*/
/*                   STATIC FUNCTIONS DEFINTIONS                              */
/*============================================================================*/

/*============================================================================*/
/* td1316_mk23_pll_set                                                         */
/*============================================================================*/
static int 
td1316_mk23_pll_set (
        struct dvb_frontend*            fe,
        struct dvb_frontend_parameters* params
        )
{
    u8 td1316_mk2_tune_data[]={0,0,0,0,0,0};
    unsigned long tunerFreq;
    unsigned long uFreqPll;
    int i = 0;
    int status_pll = 0;
    u8 buffer[4];
    struct td1316_ctx*  ctx = (struct td1316_ctx*) fe->tuner_priv;
    struct i2c_msg tuner_write_msg = {.addr = ctx->i2c_addr, .flags = 0};
    static unsigned long previous_frequency = 0;

    //Add to frequencie the IF before set the tuner PLL
    tunerFreq = params->frequency + 36166666;

    //Set the Control Data Byte
    if (tunerFreq < 180000000)
        td1316_mk2_tune_data[0] = 0xBC;
    else if (tunerFreq < 198000000)
        td1316_mk2_tune_data[0] = 0xF4;
    else if (tunerFreq < 366000000)
        td1316_mk2_tune_data[0] = 0xBC;
    else if (tunerFreq < 484000000)
        td1316_mk2_tune_data[0] = 0xF4;
    else if (tunerFreq < 700000000)
        td1316_mk2_tune_data[0] = 0xBC;
    else if (tunerFreq < 840000000)
        td1316_mk2_tune_data[0] = 0xF4;
    else    
        td1316_mk2_tune_data[0] = 0xFC;    

    //Set the Auxiliary Byte
    td1316_mk2_tune_data[0] &= 0xDF; 

    /* Calculate the divider */
    uFreqPll = tunerFreq+ (166666/2);
    uFreqPll /= 166666;

    td1316_mk2_tune_data[2] = (u8)(uFreqPll >> 8);
    td1316_mk2_tune_data[3] = (u8)uFreqPll;

    //Set the Switchport Byte
    td1316_mk2_tune_data[4] = td1316_mk2_tune_data[0] | 0x20; 

    //Auxiliary Byte
#ifdef CONFIG_BOARD_PNX8329_DTV215
    td1316_mk2_tune_data[1] = 0<<7 | 3<<4;
#else
    td1316_mk2_tune_data[1] = 1<<7 | 2<<4;	
#endif

    //Switchport Byte
    if (tunerFreq < 162000000)
        td1316_mk2_tune_data[5] = 1;
    else if (tunerFreq < 448000000)
        td1316_mk2_tune_data[5]= 2;
    else
        td1316_mk2_tune_data[5] = 4;

    //Saw filter for 8MHz
    if (params->u.ofdm.bandwidth == BANDWIDTH_8_MHZ)
        td1316_mk2_tune_data[5] |= 8;	


    tuner_write_msg.buf = buffer;
    tuner_write_msg.len = 2;
    buffer[0] = td1316_mk2_tune_data[0];
    buffer[1] = td1316_mk2_tune_data[1];
    if (i2c_transfer(ctx->i2c, &tuner_write_msg, 1) != 1)
    {
        return -EIO;
    }

    // Workaround, write control_byte & switchport_byte first
    // when the frequency is decreased
    if( tunerFreq > previous_frequency )
    {
        buffer[0] = td1316_mk2_tune_data[2];
        buffer[1] = td1316_mk2_tune_data[3];
        buffer[2] = td1316_mk2_tune_data[4];
        buffer[3] = td1316_mk2_tune_data[5];
    }
    else
    {
        buffer[0] = td1316_mk2_tune_data[4];
        buffer[1] = td1316_mk2_tune_data[5];
        buffer[2] = td1316_mk2_tune_data[2];
        buffer[3] = td1316_mk2_tune_data[3];
    }
    previous_frequency = tunerFreq;

    tuner_write_msg.buf = buffer;
    tuner_write_msg.len = 4;
    if (i2c_transfer(ctx->i2c, &tuner_write_msg, 1) != 1)
    {
        return -EIO;
    }

    msleep(1);

    i=0;
    do
    {
        status_pll = td1316_read_status(ctx);
    }while(++i < 100 && !(status_pll & 0x40));

    return 0;

}

/*============================================================================*/
/* td1316_mk14_pll_set                                                         */
/*============================================================================*/
static int 
td1316_mk14_pll_set (
        struct dvb_frontend*            fe,
        struct dvb_frontend_parameters* params
        )
{ 
    u8 td1316_mk14_tune_data[]={0,0,0,0,0,0};
    unsigned long tunerFreq;
    unsigned long uFreqPll;
    int i = 0;
    int status_pll = 0;
    struct td1316_ctx*  ctx = (struct td1316_ctx*) fe->tuner_priv;
    struct i2c_msg tuner_write_msg = {.addr = ctx->i2c_addr, .flags = 0};


    //Add to frequencie the IF before set the tuner PLL
    tunerFreq = params->frequency + 36166666;

    //-------------
    // byte 0 and 1
    //-------------
    /* Calculate the divider */
    uFreqPll = tunerFreq+ (166666/2);
    uFreqPll /= 166666;

    td1316_mk14_tune_data[0] = (u8)(uFreqPll >> 8);
    td1316_mk14_tune_data[1] = (u8)uFreqPll;

    //--------
    // byte 2
    //--------
    if (ctx->type == TDQG4)
    {
        td1316_mk14_tune_data[2] = 0x88;
    }
    if (ctx->type == DCT4C2)
    {
        // for 4C2:  DB0 DB0 CB BB CB AB
        if(tunerFreq < 488000000){
            td1316_mk14_tune_data[2] = 0x84;
        }
        else{
            td1316_mk14_tune_data[2] = 0xC4;
        }
    }
    else
    {
        td1316_mk14_tune_data[2] = 0xCA;
    }

    //--------
    // byte 3
    //--------
    // Charge pump and SAW filter section
    if (ctx->type == TDQG4)
    {
        if (tunerFreq < (153000000+IF))
            td1316_mk14_tune_data[3] = 0x45;
        else if (tunerFreq < (398000000+IF))
            td1316_mk14_tune_data[3] = 0x46;
        else if (tunerFreq < (430000000+IF))
            td1316_mk14_tune_data[3] = 0x86;
        else if (tunerFreq < (782000000+IF))
            td1316_mk14_tune_data[3] = 0x8C;
        else if (tunerFreq < (862000000+IF))
            td1316_mk14_tune_data[3] = 0xCC;
        /* ?? no default case */
    }
    else if (ctx->type == DCT4C2)
    {
        if(tunerFreq < 192000000)
            td1316_mk14_tune_data[3] = 0x09;
        else if(tunerFreq < 488000000)
            td1316_mk14_tune_data[3] = 0x0A;
        else //if(tunerFreq < 895000000)
            td1316_mk14_tune_data[3] = 0x0C;
    }
    else
    {
        if (tunerFreq < 130000000)
            td1316_mk14_tune_data[3] = 0x61;
        else if (tunerFreq < 160000000)
            td1316_mk14_tune_data[3] = 0xA1;
        else if (tunerFreq < 200000000)
            td1316_mk14_tune_data[3] = 0xC1;
        else if (tunerFreq < 290000000)
            td1316_mk14_tune_data[3] = 0x62;
        else if (tunerFreq < 420000000)
            td1316_mk14_tune_data[3] = 0xA2;
        else if (tunerFreq < 480000000)
            td1316_mk14_tune_data[3] = 0xC2;
        else if (tunerFreq < 620000000)
            td1316_mk14_tune_data[3] = 0x64;
        else if (tunerFreq < 830000000)
            td1316_mk14_tune_data[3] = 0xA4;
        else 
            td1316_mk14_tune_data[3] = 0xC4;
    }

    //Saw filter for 8MHz
    if (ctx->type == TDQG4)
    {
        if (params->u.ofdm.bandwidth == BANDWIDTH_8_MHZ)
            td1316_mk14_tune_data[3] |= 0x04;
        else
            td1316_mk14_tune_data[3] &= 0xFB;
    }
    else if (ctx->type == DCT4C2)
    {
        if (params->u.ofdm.bandwidth == BANDWIDTH_7_MHZ)
            td1316_mk14_tune_data[3] |= 0x10;
    }
    else
    {
        if (params->u.ofdm.bandwidth == BANDWIDTH_8_MHZ)
            td1316_mk14_tune_data[3] |= 0x08;
    }

    //--------
    // byte 4
    //--------
    //Auxiliary Byte
    if (ctx->type == TDQG4)
    {
        td1316_mk14_tune_data[4] = 0xE0;
    }
    if (ctx->type == DCT4C2)
    {
        // CB with AB enabled
        td1316_mk14_tune_data[4] = td1316_mk14_tune_data[2] | 0x18;
    }
    else
    {
        td1316_mk14_tune_data[4] = 0x80 | 1<<3 | 1;
    }


    //--------
    // byte 5
    //--------
    if (ctx->type == DCT4C2)
    {
        // AB
        td1316_mk14_tune_data[5] = 0x30;
    }
    else
    {
        td1316_mk14_tune_data[5] = td1316_mk14_tune_data[3];
    }

    //-------------------------------------
    // Now we send the params to the td1316
    //-------------------------------------
    tuner_write_msg.buf = td1316_mk14_tune_data;
    tuner_write_msg.len = 6;
    if (i2c_transfer(ctx->i2c, &tuner_write_msg, 1) != 1)
    {
        return -EIO;
    }

    msleep(1);

    i=0;
    do
    {
        status_pll = td1316_read_status(ctx);
    }while(++i < 100 && !(status_pll & 0x40));

    return 0;


}

/*============================================================================*/
/* td1316_read_status                                                     */
/*============================================================================*/
static int 
td1316_read_status (
        struct td1316_ctx*  ctx
        )
{
    u8 b1[]={0};
    struct i2c_msg msg[] = {{ .addr = ctx->i2c_addr, .flags = I2C_M_RD,
        .buf=b1, .len = 1}};

    if (i2c_transfer(ctx->i2c, msg, 1) != 1){
        return -EIO;
    }

    return b1[0];
}

/**
 * Standard tuner operations
 */
static struct dvb_tuner_ops tu1316_mk2_tuner_ops = {
	.info =
	{
		.name = "NXP TD1316",
		.frequency_min  = 174000000, /* TODO TBC */
		.frequency_max  = 864000000, /* TODO TBC */
		.frequency_step = 8000000
	},
  .set_params        = td1316_pll_set,
  .get_frequency     = 0,
  .get_bandwidth     = 0,
  .release           = td1316_release,
  .init              = td1316_init,
};

/*============================================================================*/
/* td1316_attach                                                          */
/*============================================================================*/
struct dvb_frontend * td1316_attach (
        struct dvb_frontend*    fe,
        u8 i2c_addr,
        struct i2c_adapter*     i2c,
        enum td1316_type        type
        )
{
    struct td1316_ctx*  ctx = NULL;

    /* allocate memory for the internal state */
    ctx = (struct td1316_ctx*)kmalloc(sizeof(struct td1316_ctx), GFP_KERNEL);
    if (ctx == NULL) {
        return NULL;
    }

    memset(ctx, 0, sizeof(struct td1316_ctx));

    ctx->i2c = i2c;
    ctx->i2c_addr = i2c_addr;
    ctx->type = type;
    
    /* type = TD1316_mkx corresponds to the classic td1316 driver for NXP tuners */
    /* type = DTT8D1T is for XuGuang DTT-8D1T tuner. Actually, DTT8D1T is completely equivalent to TD1316_mkx */
    /* type = TDQG4 is for Alps TDQG4 tuner. */
        
    memcpy(&fe->ops.tuner_ops, &tu1316_mk2_tuner_ops, sizeof(tu1316_mk2_tuner_ops));
    
    fe->tuner_priv = ctx;

    return fe;
}

EXPORT_SYMBOL(td1316_attach);
/*============================================================================*/
/*                     END OF FILE                                            */
/*============================================================================*/


