// Pandora2 SDRAM testing code.
//
// Based on the advice given at http://www.netrino.com/Articles/MemoryTesting/index.php
// and other sources.
//
// The SDRAM is arranged in four chips. The first pair of chips do the
// top and bottom sixteen bits of the first 128MB of RAM respectively
// and the second pair do the top and bottom sixteen bites of the
// second 128MB of RAM.
//
// (C) 2007 Roku LLC
//

//#include <stdio.h>
//#include <stdint.h>

#define STRIZE(x) STRIZE2(x)
#define STRIZE2(x) #x
#define assert(x) do { if (!(x)) printf(__FILE__ ":" STRIZE(__LINE__) " Assertion failed: " STRIZE(x) "\n"); } while(0)
#define countof(x) (sizeof(x)/sizeof(x[0]))

// Locked out region is where our code and data is. We sometimes
// kludge addresses in the region by adding a kludge amount to them.

#define LOCKOUT_BEGIN 	0x800000
#define LOCKOUT_END 	0x8dc000
#define LOCKOUT_KLUDGE 	0x100000

struct MemoryChip
{
    unsigned start;
    unsigned size;
};

struct MemoryChip chips[] =
{
    { 0xa0000000, 0x08000000 },
    { 0xa8000000, 0x08000000 },
};

// Because the chips split the 32-bits among themselves automatically
// an SDRAM word is 32-bits.
typedef unsigned int sdram_word;
typedef volatile sdram_word *sdram_address;

static sdram_address MakeSafeAddress(sdram_address base, unsigned long offset_in_words)
{
    sdram_address address = base + offset_in_words;
    if (((unsigned long)address >= LOCKOUT_BEGIN) && ((unsigned long)address < LOCKOUT_END))
    {
		address += LOCKOUT_KLUDGE;
    }
    if (((unsigned long)address >= LOCKOUT_BEGIN) && ((unsigned long)address < LOCKOUT_END))
    {
		assert(0);
    }
    return address;
}

static inline int NotLockedOut(sdram_address base, unsigned long offset_in_words)
{
    if (base + offset_in_words >= (sdram_address)LOCKOUT_END)
		return 1;
    if (base + offset_in_words < (sdram_address)LOCKOUT_BEGIN)
		return 1;
    return 0;
}

static unsigned int TestDataBusOnAddress(sdram_address address)
{
	unsigned failures = 0;
	
    sdram_word pattern;
    for(pattern = 1; pattern != 0; pattern <<= 1)
    {
		*address = pattern;		
		
		if (*address != pattern)
		{
			printf("*** pattern FAIL address=0x%x, value=0x%x, expected=0x%x\n", address, *address, pattern);
			++failures;
		}

		const sdram_word antipattern = ~pattern;
		
		*address = antipattern;
		
		if (*address != antipattern)
		{
			printf("*** antipattern FAIL address=0x%x, value=0x%x, expected=0x%x\n", address, *address, antipattern);
			++failures;
		}
    }
	
    return failures;
}

static unsigned int TestDataBus(void)
{
	unsigned int failures = 0;
	
    unsigned chip;
    for(chip = 0; chip < countof(chips); ++chip)
    {
		const unsigned int failure_count = TestDataBusOnAddress((sdram_address)chips[chip].start);
		if (failure_count)
		{
			printf("  DATA BUS FAIL on chip %u\n", chip);
			failures += failure_count;
		}
    }
	return failures;
}

static unsigned int TestAddressBusOnChip(sdram_address start_address, unsigned num_bytes)
{
    const sdram_word pattern = 0xa5a5a5a5;
    const sdram_word antipattern = ~pattern;
    const unsigned long address_mask = (num_bytes/sizeof(sdram_word))-1;
    
    unsigned long offset;
    unsigned long test_offset;

	unsigned int failures = 0;

    // Write default pattern to power of two offsets.
    for(offset = 1; (offset & address_mask) != 0; offset <<= 1)
    {
		sdram_address target_address = MakeSafeAddress(start_address, offset);
		*target_address = pattern;
    }

    test_offset = 0;
    *start_address = antipattern;
    
    // Check for address bits stuck high
    for(offset = 1; (offset & address_mask) != 0; offset <<= 1)
    {
		sdram_address target_address = MakeSafeAddress(start_address, offset);
		const sdram_word value = *target_address;
		if (value != pattern)
		{
			printf("*** Stuck high fail at 0x%x (read=0x%x, expected=0x%x)\n", target_address, value, pattern);
			++failures;
		}
    }
	
    *start_address = pattern;
	
    for(test_offset = 1; (test_offset & address_mask) != 0; test_offset <<= 1)
    {
		sdram_address test_target_address = MakeSafeAddress(start_address, test_offset);
		
		*test_target_address = antipattern;

		{
			const sdram_word value = *start_address;
			if (value != pattern)
			{
				printf("*** Stuck low fail at 0x%x, test_offset=0x%x (read=0x%x, expected=0x%x)\n", test_target_address, test_offset, value, pattern);
				++failures;
			}
		}
		
		for(offset = 1; (offset & address_mask) != 0; offset <<= 1)
		{
			sdram_address target_address = MakeSafeAddress(start_address, offset);
			const sdram_word value = *target_address;
			if ((value != pattern) && (offset != test_offset))
			{
				printf("*** Short fail at 0x%x read=0x%x, expected=0x%x\n", test_target_address, value, pattern);
				++failures;
			}
		}
		
		*test_target_address = pattern;
    }

	return failures;
}    

static unsigned int TestAddressBus(void)
{
	unsigned int failures = 0;
	
    unsigned chip;
    for(chip = 0; chip < countof(chips); ++chip)
    {
		unsigned int failure_count = TestAddressBusOnChip((sdram_address)chips[chip].start, chips[chip].size);
		if (failure_count != 0)
		{
			printf("  FAIL on chip pair %u: Address 0x%x\n", chip);
			failures += failure_count;
		}
    }

	// Check aliasing
	failures += TestAddressBusOnChip((sdram_address)chips[0].start, chips[0].size + chips[1].size);

	return failures;
}

unsigned int sdram_MemoryTestQuick(void)
{
	unsigned int failures = 0;
	
	printf("Testing memory: ");
	
    failures += TestDataBus();
    failures += TestAddressBus();

	printf("done\n");

	return failures;
}
